"""
Generate command

Produce the required output files for a given YAML file
"""

import os
from collections.abc import Iterable
from pathlib import Path
from typing import Optional

import click

from fgen.commands.base import cli
from fgen.data_models import (
    Package,
    PackageSharedElements,
    load_enum_defining_module,
    load_module_definition,
)
from fgen.wrapper_building import process_package


@cli.command(name="generate")
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Override any existing files",
)
@click.option(
    "--extension",
    "-e",
    help="""Name of the Python extension package""",
    required=True,
)
@click.option(
    "--wrapper-directory",
    help="Directory to write the Python-Fortran wrapper(s).",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    required=True,
)
@click.option(
    "--python-directory",
    help=("Directory in which to write the python module(s)."),
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    required=True,
)
@click.option(
    "--manager-directory",
    help="""Directory in which to write the derived type lifecycle manager(s)
    (Fortran-based).

    This defaults to the same directory as the yaml file(s)
    which define the derived type,
    which is generally where the Fortran code for the derived type is also located.
    If the YAML files come from more than one directory and this is not defined,
    an error is raised.""",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
)
@click.option(
    "--enum-definition",
    "-ed",
    help="""Yaml-file that holds the definition of an enum.

    As many of these as you like can be provided.
    Each needs to be behind its own option flag.
    For example,

    `fgen generate ... --enum-definition enum-file-1.yaml -ed enum-file-2.yaml`""",
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
    multiple=True,
)
@click.argument(
    "yaml_files",
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
    nargs=-1,
)
def generate_command(  # noqa: PLR0913
    force: bool,
    extension: str,
    wrapper_directory: str,
    python_directory: str,
    manager_directory: Optional[str],
    enum_definition: Optional[tuple[str, ...]],
    yaml_files: Iterable[str],
) -> None:
    """
    Autogenerate python and fortran code according to a collection of module definitions

    We assume that each file in YAML_FILES describes a module
    which exposes a Fortran derived type that we wish to expose to Python.
    The collection of modules is considered to be a package.

    This function writes 3 files for each module that is being wrapped:

    * manager - Lifecycle handling for derived types (in Fortran)

    * wrapper - Provides a fortran wrapper to the generated functions that can be
                made available to Python

    * python module - Python interface for creating and manipulating derived types.

    This function may also write other files necessary
    for wrapping the package as a whole.
    """
    if manager_directory is not None:
        manager_directory_to_use = manager_directory
    else:
        yaml_file_directories = tuple(os.path.dirname(fn) for fn in yaml_files)
        if len(set(yaml_file_directories)) > 1:
            msg = (
                "yaml files are in different directories. "
                "Please specify the manager directory "
                "(via the manager-directory option). "
                f"yaml file directories: {yaml_file_directories}"
            )
            raise ValueError(msg)

        manager_directory_to_use = yaml_file_directories[0]

    modules = tuple(load_module_definition(fn) for fn in yaml_files)

    if enum_definition:
        modules_enum_defining = tuple(
            load_enum_defining_module(Path(fn)) for fn in enum_definition
        )
    else:
        modules_enum_defining = ()

    package = Package(
        modules=modules,
        modules_enum_defining=modules_enum_defining,
    )
    # If we need, can make it so the initialisation of shared is configurable too.
    shared = PackageSharedElements()

    process_package(
        package,
        shared=shared,
        extension=extension,
        manager_directory=Path(manager_directory_to_use),
        wrapper_directory=Path(wrapper_directory),
        python_directory=Path(python_directory),
        force=force,
    )



"""
Exception classes for wrapper-specific errors
"""
from __future__ import annotations

from typing import Any, Callable, Optional

from numpy.typing import NDArray


class WrapperError(ValueError):
    """
    Base exception for errors that arise from wrapper
    """


class WrapperErrorUnknownCause(WrapperError):
    """
    Raised when there is an error with the wrapper, but we have nothing to help debug
    """

    def __init__(self, msg: str):
        suffix = "Underlying reason unknown"
        error_msg = f"{msg}. {suffix}."

        super().__init__(error_msg)


class InitialisationError(WrapperError):
    """
    Raised when the wrapper around the Fortran module hasn't been initialised yet
    """

    def __init__(self, instance: Any, method: Optional[Callable[..., Any]] = None):
        if method:
            error_msg = f"{instance} must be initialised before {method} is called"
        else:
            error_msg = f"instance ({instance:r}) is not initialized yet"

        super().__init__(error_msg)


class CompiledExtensionNotFoundError(ImportError):
    """
    Raised when a compiled extension can't be import i.e. found
    """

    def __init__(self, compiled_extension_name: str):
        error_msg = f"Could not find compiled extension {compiled_extension_name!r}"

        super().__init__(error_msg)


class BadPointerError(ValueError):
    """
    Raised when a pointer has a value we know is wrong
    """

    def __init__(self, pointer_value: Any, extra_info: str):
        error_msg = (
            f"The array pointer value is wrong ({pointer_value=}). "
            f"{extra_info}. "
            "Further underlying reason for the error is unknown."
        )

        super().__init__(error_msg)


class PointerArrayConversionError(ValueError):
    """
    Raised when we know a pointer has been incorrectly converted to an array
    """

    def __init__(
        self,
        array: NDArray[Any],
        ptr: int,
        expected_array_interface_data: tuple[int, bool],
    ):
        error_msg = (
            "The array pointer has been incorrectly converted to an numpy array. "
            f"We received {array.__array_interface__['data']=}. "
            f"The expected value is {expected_array_interface_data}. "
            f"Full context: {ptr=} {array.__array_interface__=}"
        )

        super().__init__(error_msg)


class UnallocatedMemoryError(ValueError):
    """
    Raised when we try to allocate memory that has not yet been allocated

    We can't always catch this error, but this is what we raise when we can.
    """

    def __init__(self, variable_name: str):
        error_msg = (
            f"The memory required to access ``{variable_name}`` is unallocated. "
            "You must allocate it before trying to access its value. "
            "Unfortunately, we cannot provide more information "
            "about why this memory is not yet allocated."
        )

        super().__init__(error_msg)


class RelativePythonModuleNotFoundError(ImportError):
    """
    Raised when a Python import configured by fgen fails
    """

    def __init__(
        self,
        requesting_python_module: str,
        requested: str,
        requested_from_python_module: str,
    ):
        """
        Initialise

        Parameters
        ----------
        requesting_python_module
            The Python module in which the ``import`` statement appears

        requested
            The thing that is being requested e.g. ``HelpfulType``

        requested_from_python_module
            The Python module from which ``requested`` is being imported e.g.
            ``source_module`` in
            ``from source_module import HelpfulType``
        """
        error_msg = (
            "There is something wrong with your fgen configuration for "
            f"{requesting_python_module!r}. "
            "Somewhere (likely in the `.yaml` file which is used to generate "
            f"{requesting_python_module!r} with fgen) you are specifying that "
            f"{requested!r} can be imported from "
            f"{requested_from_python_module!r}, but this is failing."
        )

        super().__init__(error_msg)


class SolveError(ValueError):
    """
    Exception raised when an solver can't solve

    For example, because the integration has a runaway effect in it
    """



"""
Data model of a module that defines an enum
"""
from __future__ import annotations

from attrs import define


@define
class EnumValue:
    """
    Data model of an enum value
    """

    description: str
    """Description of the value"""

    str_value: str
    """The string value of the enum."""

    integer_value: int
    """
    The integer value of the enum.

    Ensures consistency across Python and Fortran.
    """


@define
class EnumDefinition:
    """
    Data model of an enum definition
    """

    name: str
    """Name of the enum (in both Fortran and Python)"""

    description: str
    """Description of enum"""

    values: tuple[EnumValue, ...]
    """Values the enum can take"""


@define
class ModuleEnumDefining:
    """
    Data model of a module that defines an enum
    """

    name: str
    """Name of the module"""

    description: str
    """Description of the module"""

    provides: EnumDefinition
    """Enum that this module defines"""



"""
Formatting of fgen wrapped objects
"""
from __future__ import annotations

from collections.abc import Iterable
from typing import Any

import fgen_runtime.exceptions as fgr_excs
from fgen_runtime.base import INVALID_INSTANCE_INDEX, FinalizableWrapperBase


def get_attribute_str_value(instance: FinalizableWrapperBase, attribute: str) -> str:
    """
    Get the string version of an attribute's value

    Parameters
    ----------
    instance
        Instance from which to get the attribute

    attribute
        Attribute for which to get the value

    Returns
    -------
        String version of the attribute's value, with graceful handling of errors.
    """
    try:
        return f"{attribute}={getattr(instance, attribute)}"
    except fgr_excs.UnallocatedMemoryError:
        return f"{attribute} is unallocated"


def to_str(instance: FinalizableWrapperBase, exposed_attributes: Iterable[str]) -> str:
    """
    Convert an instance to its string representation

    Parameters
    ----------
    instance
        Instance to convert

    exposed_attributes
        Attributes from Fortran that the instance exposes

    Returns
    -------
        String representation of the instance
    """
    if instance.instance_index == INVALID_INSTANCE_INDEX:
        return f"Uninitialised {instance!r}"

    if not exposed_attributes:
        return repr(instance)

    attribute_values = [
        get_attribute_str_value(instance, v) for v in exposed_attributes
    ]

    return f"{repr(instance)[:-1]}, {', '.join(attribute_values)})"


def to_pretty(
    instance: FinalizableWrapperBase,
    exposed_attributes: Iterable[str],
    p: Any,
    cycle: bool,
    indent: int = 4,
) -> None:
    """
    Pretty-print an instance

    Parameters
    ----------
    instance
        Instance to convert

    exposed_attributes
        Attributes from Fortran that the instance exposes

    p
        Pretty printing object

    cycle
        Whether the pretty printer has detected a cycle or not.

    indent
        Indent to apply to the pretty printing group
    """
    if instance.instance_index == INVALID_INSTANCE_INDEX:
        p.text(str(instance))
        return

    if not exposed_attributes:
        p.text(str(instance))
        return

    with p.group(indent, f"{repr(instance)[:-1]}", ")"):
        for att in exposed_attributes:
            p.text(",")
            p.breakable()

            p.text(get_attribute_str_value(instance, att))


def add_attribute_row(
    attribute_name: str, attribute_value: str, attribute_rows: list[str]
) -> list[str]:
    """
    Add a row for displaying an attribute's value to a list of rows

    Parameters
    ----------
    attribute_name
        Attribute's name

    attribute_value
        Attribute's value

    attribute_rows
        Existing attribute rows


    Returns
    -------
        Attribute rows, with the new row appended
    """
    attribute_rows.append(
        f"<tr><th>{attribute_name}</th><td style='text-align:left;'>{attribute_value}</td></tr>"  # noqa: E501
    )

    return attribute_rows


def to_html(instance: FinalizableWrapperBase, exposed_attributes: Iterable[str]) -> str:
    """
    Convert an instance to its html representation

    Parameters
    ----------
    instance
        Instance to convert

    exposed_attributes
        Attributes from Fortran that the instance exposes

    Returns
    -------
        HTML representation of the instance
    """
    if instance.instance_index == INVALID_INSTANCE_INDEX:
        return str(instance)

    if not exposed_attributes:
        return str(instance)

    instance_class_name = repr(instance).split("(")[0]

    attribute_rows: list[str] = []
    for att in exposed_attributes:
        try:
            att_val = getattr(instance, att)
        except fgr_excs.UnallocatedMemoryError:
            att_val = "Unallocated"
            attribute_rows = add_attribute_row(att, att_val, attribute_rows)
            continue

        try:
            att_val = att_val._repr_html_()
        except AttributeError:
            att_val = str(att_val)

        attribute_rows = add_attribute_row(att, att_val, attribute_rows)

    attribute_rows_for_table = "\n          ".join(attribute_rows)

    css_style = """.fgen-wrap {
  /*font-family: monospace;*/
  width: 540px;
}

.fgen-header {
  padding: 6px 0 6px 3px;
  border-bottom: solid 1px #777;
  color: #555;;
}

.fgen-header > div {
  display: inline;
  margin-top: 0;
  margin-bottom: 0;
}

.fgen-basefinalizable-cls,
.fgen-basefinalizable-instance-index {
  margin-left: 2px;
  margin-right: 10px;
}

.fgen-basefinalizable-cls {
  font-weight: bold;
  color: #000000;
}"""

    return "\n".join(
        [
            "<div>",
            "  <style>",
            f"{css_style}",
            "  </style>",
            "  <div class='fgen-wrap'>",
            "    <div class='fgen-header'>",
            f"      <div class='fgen-basefinalizable-cls'>{instance_class_name}</div>",
            f"        <div class='fgen-basefinalizable-instance-index'>instance_index={instance.instance_index}</div>",  # noqa: E501
            "        <table><tbody>",
            f"          {attribute_rows_for_table}",
            "        </tbody></table>",
            "    </div>",
            "  </div>",
            "</div>",
        ]
    )



"""
Generation of the Fortran wrapper module
"""
from __future__ import annotations

from collections.abc import Iterable
from pathlib import Path
from typing import Union

from attrs import define, evolve, frozen

from fgen.data_models import Module, MultiReturn, Package, PackageSharedElements, Value
from fgen.jinja_environment import (
    JINJA_ENV,
    get_template_in_directory,
    post_process_jinja_rendering,
)
from fgen.wrapping_strategies import (
    WrappingStrategyLike,
    get_wrapping_strategy,
)
from fgen.wrapping_strategies.receiving_from_python_steps import (
    ReceivingFromPythonSteps,
)


@define
class FortranWrapperModuleBuilder:
    """
    Builder of a Fortran wrapper module
    """

    package: Package
    """
    Package for which the builder is building wrappers
    """

    module: Module
    """
    Module for which to build the wrapper
    """

    shared: PackageSharedElements
    """
    Elements which have to be shared across the package

    For example, the names of functions which are used in more than one wrapper module.
    """

    def get_wrapping_strategy(
        self, value: Union[MultiReturn, Value]
    ) -> WrappingStrategyLike:
        """
        Get wrapping strategy

        Parameters
        ----------
        value
            Value for which to get the wrapping strategy

        Returns
        -------
            Wrapping strategy to use with the value
        """
        return get_wrapping_strategy(value.definition.as_fortran_data_type())

    def get_module_level_docstring(self) -> str:
        """
        Get the module-level docstring

        Returns
        -------
            Module-level docstring
        """
        template = get_template_in_directory(
            "module-level-docstring.f90.jinja", Path(__file__).parent, JINJA_ENV
        )

        result = template.render(module=self.module)

        return result

    def get_module_use_statements(self) -> str:
        """
        Get the module-level use statements

        Returns
        -------
            Module-level use statements
        """
        template = get_template_in_directory(
            "module-use-statements.f90.jinja", Path(__file__).parent, JINJA_ENV
        )

        result = template.render(builder=self)

        return result

    def get_requirements_fortran_use_statements(self) -> str:
        """
        Get the Fortran ``use`` statements needed for the module

        Returns
        -------
            Fortran ``use`` statements needed for the module
        """
        requirements_template = get_template_in_directory(
            "module-use-requirements.f90.jinja", Path(__file__).parent, JINJA_ENV
        )

        required_use_statements = self.get_requirements_fortran_uses(self.module)

        out: list[str] = []
        for providing_module_name in sorted(required_use_statements.keys()):
            needed_type = required_use_statements[providing_module_name]

            if isinstance(needed_type.providing_module, Module):
                providing_module_manager_name = (
                    needed_type.providing_module.manager_module_name
                )
            else:
                providing_module_manager_name = needed_type.providing_module

            rendered = requirements_template.render(
                providing_module_name=providing_module_name,
                required_type=needed_type.name,
                requires_get_instance=needed_type.requires_get_instance,
                requires_get_free_instance_number=needed_type.requires_get_free_instance_number,
                providing_module_manager_name=providing_module_manager_name,
                shared=self.shared,
            )
            # Strip out blank and trailing new lines,
            # not sure how to make jinja behave so just using Python instead.
            out.append("\n".join([v for v in rendered.splitlines()]))

        return "\n".join(out)

    def get_requirements_fortran_uses(self, module: Module) -> dict[str, NeededType]:
        """
        Get the Fortran ``use``'s needed for the module

        Parameters
        ----------
        module
            Module for which to get the ``use``'s

        Returns
        -------
            Fortran ``use`` statements needed for the module
        """
        required_use_statements: dict[str, NeededType] = {}

        provided_type = module.provides
        provided_type_attributes = provided_type.exposed_attributes.values()
        provided_type_method_arguments = [
            argument
            for method in provided_type.methods.values()
            for argument in method.parameters.values()
        ]
        provided_type_method_return_values = [
            method.returns for method in provided_type.methods.values()
        ]
        for value, include_get_free_instance_number in [
            *[(v, True) for v in provided_type_attributes],
            *[(v, False) for v in provided_type_method_arguments],
            *[(v, True) for v in provided_type_method_return_values],
        ]:
            if self.requires_import(value, module):
                needed_type_str = value.definition.python_type_as_str()
                fdt = value.definition.as_fortran_data_type()
                if fdt.is_enum:
                    sought_type = fdt.equivalent_python_type
                    for requirement in module.requirements:
                        if requirement.provides == sought_type:
                            break
                    else:
                        msg = (
                            "Please specify where to get the enum "
                            f"{sought_type} from in `module.requirements`"
                            f"for {module=}"
                        )
                        ValueError(msg)

                else:
                    requirement = module.get_requirement_that_provides(needed_type_str)

                if (
                    not fdt.is_enum
                    and requirement.fortran_module in required_use_statements
                ):
                    if (
                        needed_type_str
                        != required_use_statements[requirement.fortran_module].name
                    ):
                        msg = (
                            "Each module should only provide one type. "
                            f"You are requesting {needed_type_str}, "
                            f"but have already said that {requirement} "
                            "provides "
                            f"{required_use_statements[requirement.fortran_module]}"
                        )
                        raise AssertionError(msg)

                    updated_requires_get_free_instance_number = (
                        include_get_free_instance_number
                        or required_use_statements[
                            requirement.fortran_module
                        ].requires_get_free_instance_number
                    )

                    required_use_statements[requirement.fortran_module] = evolve(
                        required_use_statements[requirement.fortran_module],
                        requires_get_free_instance_number=updated_requires_get_free_instance_number,
                    )

                else:
                    if fdt.is_enum:
                        needed_type = NeededType(
                            name=needed_type_str,
                            providing_module=requirement.fortran_module,
                            requires_get_instance=False,
                            requires_get_free_instance_number=False,
                        )

                    else:
                        requirement_providing_module = (
                            self.package.get_module_that_provides_values_type(value)
                        )
                        needed_type = NeededType(
                            name=needed_type_str,
                            providing_module=requirement_providing_module,
                            requires_get_instance=True,
                            requires_get_free_instance_number=include_get_free_instance_number,
                        )

                    required_use_statements[requirement.fortran_module] = needed_type

        return required_use_statements

    def requires_import(self, value: Union[Value, MultiReturn], module: Module) -> bool:
        """
        Return whether a value requires an import or not

        Parameters
        ----------
        value
            Value to check

        module
            Module in which the value is being used

        Returns
        -------
            ``True`` if this values requires an import, ``False`` otherwise.
        """
        fdt = value.definition.as_fortran_data_type()

        if fdt.is_derived_type:
            # If the type we need is defined by the module, we don't need to import it
            return not (fdt.equivalent_python_type == module.provides.name)

        if fdt.is_enum:
            return True

        return False

    def get_statement_declarations_getters_and_setters(self) -> str:
        """
        Get the statement declarations for getters and setters

        Returns
        -------
            Statement declarations for the getters and setters
        """
        out: list[str] = []
        for attribute in self.module.provides.exposed_attributes.values():
            ws = self.get_wrapping_strategy(attribute)

            statement_declarations = (
                ws.get_fortran_wrapper_statement_declarations_getters_and_setters(
                    attribute
                )
            )
            out.extend(statement_declarations)

        return "\n".join(out)

    def get_statement_declarations_methods(self) -> str:
        """
        Get the statement declarations for methods

        Returns
        -------
            Statement declarations for the methods
            exposed by the derived type defined by this module.
        """
        out: list[str] = []
        for method in self.module.provides.methods.values():
            return_value = method.returns
            ws = self.get_wrapping_strategy(return_value)

            statement_declarations = ws.get_fortran_wrapper_statement_declarations_for_returning_method_result_to_python(  # noqa: E501
                method, prefix=self.shared.fortran_wrapper_method_prefix
            )
            out.extend(statement_declarations)

        return "\n".join(out)

    def get_provided_type_build_methods(self) -> str:
        """
        Get the provided type's build methods

        Returns
        -------
            Class methods
        """
        template = get_template_in_directory(
            "provided-type-method-no-return.f90.jinja", Path(__file__).parent, JINJA_ENV
        )

        receiving_from_python_steps = self.get_receiving_from_python_steps(
            self.module.provides.attributes.values()
        )
        input_params = (
            receiving_from_python_steps.fortran_wrapper_callable_parameter_names
        )

        result = template.render(
            builder=self,
            shared=self.shared,
            fortran_wrapper_callable=self.shared.fortran_build_callable_name,
            fortran_module_callable="instance % build",
            receiving_from_python_steps=receiving_from_python_steps,
            input_params=input_params,
        )

        return result

    def get_provided_type_getters_and_setters(self) -> str:
        """
        Get the provided type's getters and setters

        Returns
        -------
            Provided type's getters and setters
        """
        template_setter = get_template_in_directory(
            "provided-type-setter.f90.jinja",
            Path(__file__).parent,
            JINJA_ENV,
        )
        result: list[str] = []

        for att in self.module.provides.exposed_attributes.values():
            att_ws = self.get_wrapping_strategy(att)

            if att.definition.as_fortran_data_type().is_derived_type:
                att_providing_module = (
                    self.package.get_module_that_provides_values_type(att)
                )
                att_manager_get_free_instance_number = "_".join(
                    [
                        att_providing_module.manager_module_name,
                        self.shared.fortran_get_free_instance_number_callable_name,
                    ]
                )
                att_manager_get_instance = "_".join(
                    [
                        att_providing_module.manager_module_name,
                        self.shared.fortran_get_instance_callable_name,
                    ]
                )
            else:
                att_manager_get_free_instance_number = None
                att_manager_get_instance = None

            att_getter = att_ws.get_fortran_for_getter(
                value=att,
                class_being_wrapped=self.module.provides.name,
                value_manager_get_free_instance_number=att_manager_get_free_instance_number,
                value_manager_get_instance=att_manager_get_instance,
            )
            result.append(att_getter)

            receiving_from_python_steps = self.get_receiving_from_python_steps([att])
            if len(receiving_from_python_steps.fortran_module_callable_args) != 1:
                msg = (
                    "Only one value should be needed for the setter. We have: "
                    f"{receiving_from_python_steps.fortran_module_callable_args=}"
                )
                raise AssertionError(msg)

            fortran_module_attribute_to_set = (
                receiving_from_python_steps.fortran_module_callable_args[0]
            )
            setter_input_params = (
                receiving_from_python_steps.fortran_wrapper_callable_parameter_names
            )

            att_setter = template_setter.render(
                builder=self,
                fortran_wrapper_callable=att_ws.get_fortran_wrapper_module_setter(att),
                receiving_from_python_steps=receiving_from_python_steps,
                input_params=setter_input_params,
                fortran_module_attribute_to_set=fortran_module_attribute_to_set,
            )
            result.append(att_setter)

        return "\n\n".join(result)

    def get_provided_type_methods(self) -> str:
        """
        Get the provided type's methods

        Returns
        -------
            Provided type's methods
        """
        result: list[str] = []
        for method in self.module.provides.methods.values():
            method_returns = method.returns
            ws_method_returns = self.get_wrapping_strategy(method_returns)

            if method_returns.definition.as_fortran_data_type().is_derived_type:
                method_returns_providing_module = (
                    self.package.get_module_that_provides_values_type(method_returns)
                )

                if method_returns_providing_module == self.module:
                    # TODO: remove hard-coding here and use shared instead.
                    # Also requires changes to module-use-statements.f90.jina.
                    method_returns_manager_get_free_instance_number = (
                        "manager_get_free_instance"
                    )
                    method_returns_manager_get_instance = "manager_get_instance"

                else:
                    method_returns_manager_get_free_instance_number = "_".join(
                        [
                            method_returns_providing_module.manager_module_name,
                            self.shared.fortran_get_free_instance_number_callable_name,
                        ]
                    )
                    method_returns_manager_get_instance = "_".join(
                        [
                            method_returns_providing_module.manager_module_name,
                            self.shared.fortran_get_instance_callable_name,
                        ]
                    )

            else:
                method_returns_manager_get_free_instance_number = None
                method_returns_manager_get_instance = None

            receiving_from_python_steps = self.get_receiving_from_python_steps(
                method.parameters.values()
            )
            method_fortran = ws_method_returns.get_fortran_for_method_returning_wrapped_type(  # noqa: E501
                fortran_wrapper_module_callable=f"{self.shared.fortran_wrapper_method_prefix}{method.name}",
                receiving_from_python_steps=receiving_from_python_steps,
                return_value=method_returns,
                class_being_wrapped=self.module.provides.name,
                method_name=method.name,
                return_value_manager_get_free_instance_number=method_returns_manager_get_free_instance_number,
                return_value_manager_get_instance=method_returns_manager_get_instance,
            )
            result.append(method_fortran)

        return "\n\n".join(result)

    def get_receiving_from_python_steps(
        self,
        arguments: Iterable[Union[MultiReturn, Value]],
    ) -> ReceivingFromPythonSteps:
        """
        Get receiving from Python steps for a set of arguments

        Parameters
        ----------
        arguments
            Arguments for which to get the receiving from Python steps

        Returns
        -------
            Steps for receiving the arguments from Python
        """
        postprocessing_fortran_calls_list = []
        fortran_wrapper_callable_parameters: list[
            tuple[str, str, Union[MultiReturn, Value]]
        ] = []
        fortran_module_callable_args: list[Union[MultiReturn, Value]] = []

        for argument in arguments:
            ws = self.get_wrapping_strategy(argument)

            argument_steps = ws.get_receiving_from_python_steps(
                value=argument,
                providing_module=self.package.find_providing_module(argument),
                get_instance_callable_name=self.shared.fortran_get_instance_callable_name,
                # At the moment, dynamic unit doesn't affect
                # the name on the Fortran side
                # so we can simply always pass None here.
                dynamic_unit=None,
            )

            if argument_steps.postprocessing_fortran_calls is not None:
                postprocessing_fortran_calls_list.append(
                    argument_steps.postprocessing_fortran_calls
                )

            fortran_wrapper_callable_parameters.extend(
                argument_steps.fortran_wrapper_callable_parameters
            )

            fortran_module_callable_args.extend(
                argument_steps.fortran_module_callable_args
            )

        if postprocessing_fortran_calls_list:
            postprocessing_fortran_calls = "\n".join(postprocessing_fortran_calls_list)
        else:
            postprocessing_fortran_calls = None

        return ReceivingFromPythonSteps(
            postprocessing_fortran_calls=postprocessing_fortran_calls,
            fortran_wrapper_callable_parameters=tuple(
                fortran_wrapper_callable_parameters
            ),
            fortran_module_callable_args=tuple(fortran_module_callable_args),
        )


@frozen
class NeededType:
    """
    Helper class for keeping track of the first-party types needed by the modle
    """

    name: str
    """Name of the needed type"""

    providing_module: Union[Module, str]
    """Module or name of the module which provides this type"""

    requires_get_instance: bool
    """
    Whether this needed type also requires its ``get_instance`` function
    """

    requires_get_free_instance_number: bool
    """
    Whether this needed type also requires its ``get_free_instance_number`` function
    """


def generate_fortran_wrapper_module(builder: FortranWrapperModuleBuilder) -> str:
    """
    Generate the Fortran wrapper module

    Parameters
    ----------
    builder
        Builder to use to generate the Fortran wrapper module

    Returns
    -------
        Fortran wrapper module as code
    """
    template = get_template_in_directory(
        "fortran-wrapper-module.f90.jinja", Path(__file__).parent, JINJA_ENV
    )

    result = post_process_jinja_rendering(
        # TODO: switch to passing in builder
        template.render(builder=builder, package=builder.package, module=builder.module)
    )

    return result



"""
Wrapping strategy for characters with a deferred size (i.e. length)

These are tricky because you have to know the length of the character
when retrieving it from the Python side for the f2py wrappers to build.
"""

from __future__ import annotations

from collections.abc import Iterable
from pathlib import Path
from typing import Optional, Union

from attrs import define

from fgen.data_models import Method, Module, MultiReturn, Value
from fgen.jinja_environment import (
    JINJA_ENV,
    get_template_in_directory,
)
from fgen.wrapping_strategies.information_injection import (
    inject_wrapping_strategy_information,
)
from fgen.wrapping_strategies.passing_to_fortran_steps import (
    PassingToFortranSteps,
)
from fgen.wrapping_strategies.receiving_from_python_steps import (
    ReceivingFromPythonSteps,
)


@define
class WrappingStrategyCharacterDeferredSize:
    """
    Wrapping strategy for characters with a deferred size (i.e. length)
    """

    length_suffix: str = "_length"
    """Suffix to use for variables that represent the length of the character"""

    length_callable_suffix: str = "_length"
    """Suffix to use for callables that return the length of the character"""

    def get_fortran_wrapper_module_getter(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        return f"iget_{value.definition.name}"

    def get_fortran_wrapper_module_setter(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        return f"iset_{value.definition.name}"

    def get_fortran_wrapper_statement_declarations_getters_and_setters(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> list[str]:
        template = get_template_in_directory(
            "fortran-statement-declarations-getter-setter.f90.jinja",
            Path(__file__).parent,
            JINJA_ENV,
        )

        result = template.render(
            getter_name=self.get_fortran_wrapper_module_getter(value),
            length_callable_suffix=self.length_callable_suffix,
            setter_name=self.get_fortran_wrapper_module_setter(value),
        )

        return result.splitlines()

    def get_fortran_wrapper_statement_declarations_for_returning_method_result_to_python(  # noqa: E501, D102
        self, method: Method, prefix: str
    ) -> list[str]:
        template = get_template_in_directory(
            "fortran-statement-declarations-method-returning-to-python.f90.jinja",
            Path(__file__).parent,
            JINJA_ENV,
        )

        result = template.render(
            method_name=method.name,
            prefix=prefix,
            length_callable_suffix=self.length_callable_suffix,
        )

        return result.splitlines()

    def get_receiving_from_python_steps(  # noqa: D102
        self,
        value: Union[MultiReturn, Value],
        dynamic_unit: Optional[str] = None,
        providing_module: Optional[Module] = None,
        get_instance_callable_name: Optional[str] = None,
    ) -> ReceivingFromPythonSteps:
        return ReceivingFromPythonSteps(
            postprocessing_fortran_calls=None,
            fortran_wrapper_callable_parameters=(
                (
                    value.definition.name,
                    # When receiving from Python,
                    # we can use an assumed length string
                    "character(len=*)",
                    value,
                ),
            ),
            fortran_module_callable_args=(value,),
        )

    def get_type_attribute_declaration_for_value_to_pass_to_original_fortran_module(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        # If intermediate variables are required, they are handled as assumed length
        ftad = (
            value.definition.as_fortran_data_type().fortran_type_attribute_declaration
        )
        return ftad.replace(":", "*")

    def get_fortran_for_getter(  # noqa: D102
        self,
        value: Union[MultiReturn, Value],
        class_being_wrapped: str,
        value_manager_get_free_instance_number: Optional[str] = None,
        value_manager_get_instance: Optional[str] = None,
    ) -> str:
        template = get_template_in_directory(
            "fortran-for-getter.f90.jinja",
            Path(__file__).parent,
            JINJA_ENV,
        )

        result = template.render(
            return_value=value,
            fortran_wrapper_module_callable=self.get_fortran_wrapper_module_getter(
                value
            ),
            class_being_wrapped=class_being_wrapped,
            fortran_module_attribute=value.definition.name,
            length_suffix=self.length_suffix,
            length_callable_suffix=self.length_callable_suffix,
        )

        return inject_wrapping_strategy_information(
            result, value=value, wrapping_strategy=self, comment_character="!"
        )

    def get_fortran_for_method_returning_wrapped_type(  # noqa: D102, PLR0913
        self,
        fortran_wrapper_module_callable: str,
        receiving_from_python_steps: ReceivingFromPythonSteps,
        return_value: Union[MultiReturn, Value],
        class_being_wrapped: str,
        method_name: str,
        return_value_manager_get_free_instance_number: Optional[str],
        return_value_manager_get_instance: Optional[str],
    ) -> str:
        template = get_template_in_directory(
            "fortran-for-method-returning-wrapped-type.f90.jinja",
            Path(__file__).parent,
            JINJA_ENV,
        )

        input_params = (
            receiving_from_python_steps.fortran_wrapper_callable_parameter_names
        )

        result = template.render(
            fortran_wrapper_module_callable=fortran_wrapper_module_callable,
            return_value=return_value,
            input_params=input_params,
            receiving_from_python_steps=receiving_from_python_steps,
            class_being_wrapped=class_being_wrapped,
            method_name=method_name,
            length_suffix=self.length_suffix,
            length_callable_suffix=self.length_callable_suffix,
        )

        return inject_wrapping_strategy_information(
            result, value=return_value, wrapping_strategy=self, comment_character="!"
        )

    def get_python_user_facing_name(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        return value.definition.name

    def get_python_return_type_annotation(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        return value.definition.as_fortran_data_type().python_equivalent_type_annotation

    def get_python_argument_declaration_type_annotation(  # noqa: D102
        self, value: Value
    ) -> str:
        return value.definition.as_fortran_data_type().python_equivalent_type_annotation

    def get_python_getter_docstring(self, value: Value) -> str:  # noqa: D102
        template = get_template_in_directory(
            "python-getter-docstring.py.jinja", Path(__file__).parent, JINJA_ENV
        )

        result = template.render(
            attribute=value,
            is_pointer=value.definition.as_fortran_data_type().is_pointer,
        )

        return result

    def get_python_setter_docstring(self, value: Value) -> str:  # noqa: D102
        return f"Setter for ``{value.definition.name}``"

    def get_python_post_verify_units_input_type_annotation(  # noqa: D102
        self, value: Union[MultiReturn, Value]
    ) -> str:
        return value.definition.as_fortran_data_type().python_equivalent_type_annotation

    def generate_python_for_fortran_return_value_processing(  # noqa: D102
        self,
        value: Union[MultiReturn, Value],
        fortran_module_callable: str,
        fortran_module_callable_args: Iterable[str],
        dynamic_unit: Optional[str],
    ) -> str:
        return_type = self.get_python_return_type_annotation(value)

        template = get_template_in_directory(
            "python-for-fortran-return-value-processing.py.jinja",
            Path(__file__).parent,
            JINJA_ENV,
        )

        result = template.render(
            result_name=self.get_python_user_facing_name(value),
            fortran_module_callable=fortran_module_callable,
            fortran_module_callable_args=fortran_module_callable_args,
            length_callable_suffix=self.length_callable_suffix,
            length_suffix=self.length_suffix,
            return_type=return_type,
        )

        return inject_wrapping_strategy_information(
            result, value=value, wrapping_strategy=self, comment_character="#"
        )

    def get_passing_to_fortran_steps(  # noqa: D102
        self,
        value: Union[MultiReturn, Value],
        dynamic_unit: Optional[str],
    ) -> PassingToFortranSteps:
        return PassingToFortranSteps(
            preparatory_python_calls=None,
            fortran_module_callable_args=(
                (value.definition.name, self.get_python_user_facing_name(value)),
            ),
        )



"""
Data model of a value (e.g. parameter, single return value)
"""
from __future__ import annotations

from typing import Any, Optional, TypeVar, Union

from attrs import define, field

from fgen.data_models.unitless_value import UnitlessValue

T = TypeVar("T")


def no_conversion(inp: T) -> T:
    """
    Identity function, just returns what it is given i.e. performs no conversion

    The use case for this is disabling type conversion with :mod:`cattrs`
    and :mod:`attrs`. In order to get :mod:`cattrs` to not convert stuff,
    you have to tell the converter a) to prefer attrs converters (i.e. use
    ``prefer_attrib_converters=True`` when creating your :mod:`cattrs`
    converter) and b) actual supply a converter to your :mod:`attrs`
    attribute. If you don't actually want to do any conversion, then you
    need an identity function (otherwise, :mod:`cattrs` will assume you
    didn't think about conversion and inject its own conversion). This
    lets the validator handle type checking etc., rather than having that
    check happen in a separate spot (without any context about what is
    being checked etc.).

    Parameters
    ----------
    inp
        Input

    Returns
    -------
        ``inp``, unchanged.
    """
    return inp


@define
class Value:
    """
    Data model of a value

    This defines the value's unit, Fortran data type and other metadata.
    It is the combination of a :obj:`UnitlessValueDefinition` and unit information.
    """

    definition: UnitlessValue
    """Definition of the value's key information"""

    unit: Optional[str] = field(default=None, converter=no_conversion)
    """
    Unit of the value

    The unit must be able to parsed by `pint`
    and be present in the :obj:`pint.UnitRegistry`
    being used by the application (normal rules for pint).
    Some examples include: "kg", "1 / month".

    A unit is required for all numeric-types (i.e. integer, real, complex).
    For non-numeric values, the unit is unused.
    """

    dynamic_unit: Union[bool, str] = False
    """
    Whether the unit should be inferred dynamically, rather than statically

    If this is ``True``, we will infer the unit
    using the units of passed :obj:`pint.Quantity`'s.
    When passing these values to Fortran,
    the unit will be extracted and passed to Fortran as a string
    to the attribute whose
    :py:attr:`~fgen.data_models.unitless_value.UnitlessValue.is_fortran_units_holder`
    is ``True``.
    If :attr:`dynamic_unit` is ``True``, when retrieving the values from Fortran,
    the unit will be requested from Fortran too
    (from the attribute whose
    :py:attr:`~fgen.data_models.unitless_value.UnitlessValue.is_fortran_units_holder`
    is ``True``)
    and added to the return value to make a :obj:`pint.Quantity` before returning.

    If this is a string, we assume this tells us where to retrieve the unit information
    from on the Python side (i.e. the string should be valid Python code).
    """

    @unit.validator
    def _check_unit(self, attribute: Any, value: str | None) -> None:
        if self.requires_units:
            if value is None:
                raise ValueError(  # noqa: TRY003
                    f"A unit is required for: {self.definition.name}"
                )

            if not isinstance(value, str):
                raise TypeError(  # noqa: TRY003
                    f"The unit for {self.definition.name} must be a string, "
                    f"received: {value}"
                )

    @property
    def requires_units(self) -> bool:
        """
        Whether this value requires units or not

        Returns
        -------
            ``True`` if this value requires units, ``False`` otherwise.
        """
        # TODO: loosen this. Not all numeric types require units
        # (e.g. integers that represent sizes of things).
        return (not self.dynamic_unit) and self.definition.is_numeric_type



"""
Functions to use to format generated code
"""
from __future__ import annotations

import black


def format_python_code(code: str) -> str:
    """
    Format Python code

    Parameters
    ----------
    code
        Code to format

    Returns
    -------
        Formatted Python code
    """
    # Format result with black
    # This doesn't automatically read `pyproject.toml` so settings must be prescribed
    # TODO: update to formatting with ruff to avoid formatters fighting each other
    return black.format_str(code, mode=black.Mode(line_length=88))



"""
Handle the translation of a fortran pointer into a numpy array
"""
import functools
import re
from typing import Any

import numpy as np
from numpy.typing import NDArray

from fgen_runtime.exceptions import PointerArrayConversionError

KIND_REGEX = r"\((?:kind=)?([1-8])\)"
"""Regex for extracting the kind information from a type specification

Only supports real, integer, logical and double precision types where "kind" is the only
available argument.
"""


def _ctype_ndarray(element_type: Any, shape: tuple[int, ...]) -> Any:
    """
    Create a ndarray of the given element type and shape

    This took some inspiration from np.ctypeslib
    """
    for dim in shape:
        element_type = dim * element_type
        # prevent the type name include np.ctypeslib
        element_type.__module__ = None
    return element_type


def _ptr_to_ndarray(ptr: int, dtype: str, shape: tuple[int, ...]) -> NDArray[Any]:
    """
    Convert a c pointer for a fortran array into a numpy array

    Any changes made to the numpy array will change the same memory used by fortran
    """
    ctype_scalar = np.ctypeslib.as_ctypes_type(dtype)
    result_type = _ctype_ndarray(ctype_scalar, shape)
    result = result_type.from_address(ptr)

    d = np.ascontiguousarray(result).T

    exp = (ptr, False)
    if d.__array_interface__["data"] != exp:
        raise PointerArrayConversionError(
            array=d, ptr=ptr, expected_array_interface_data=exp
        )

    return d


@functools.lru_cache(128)
def ftype_to_np(type_specification: str) -> str:
    """
    Convert a fortran type specification into a numpy dtype

    Only supports logical, integer and real types.

    Parameters
    ----------
    type_specification
        Type specification

        See :mod:`fgen.fortran_parsing`
        for more information about fortran type specifications.

    Returns
    -------
        A numpy equivalent dtype (where possible)

    Raises
    ------
    ValueError
        If the dtype cannot be parsed
    """
    supported_types = {"real": "f", "integer": "i", "logical": "b"}
    # Assumes no additional compiler flags are specified
    # https://gcc.gnu.org/onlinedocs/gfortran/KIND-Type-Parameters.html
    default_kind = 4

    tokens = type_specification.split("(")

    try:
        f_type = tokens[0]

        np_type = supported_types[f_type]
    except KeyError as exc:
        raise ValueError(  # noqa: TRY003
            f"Unsupported type: {type_specification}"
        ) from exc

    if len(tokens) > 1:
        match = re.search(
            # Strip out any whitespace before checking
            KIND_REGEX,
            type_specification.replace(" ", ""),
            flags=re.IGNORECASE,
        )
        if match:
            kind = match.group(1)
        else:
            raise ValueError("Could not determine kind information")  # noqa: TRY003

        if f_type == "real" and int(kind) not in [4, 8]:
            raise ValueError("real only supports 4 and 8 byte types")  # noqa: TRY003
    else:
        kind = default_kind

    return f"{np_type}{kind}"



"""
Jinja environment to use while generating wrappers
"""

from __future__ import annotations

from pathlib import Path

import jinja2

JINJA_ENV = jinja2.Environment(
    loader=jinja2.FileSystemLoader(Path(__file__).parent.absolute()),
    autoescape=jinja2.select_autoescape(),
    trim_blocks=True,
    lstrip_blocks=True,
    undefined=jinja2.StrictUndefined,
)
"""Jinja2 environment to use for template rendering"""


def indent_based_on_first_line(
    code: str,
    indent: str,
) -> str:
    """
    Indent code based on the indent of the first line.

    If the first line does not have indent equal to ``indent``,
    all the code except the first line
    will have an indent added
    such that the first line would have an indent of ``indent``.
    We exclude the first line
    so that the indent can be included in the jinja template too,
    which makes it easier to read what is going on.

    Parameters
    ----------
    code
        Code to indent

    indent
        Indent to ensure that the first line has


    Returns
    -------
        Indented code
    """
    if not code:
        # Empty line
        return code

    code_split = code.splitlines()

    if len(code_split) == 1:
        # Nothing to indent
        return code

    first_line = code_split[0]

    first_line_indent = len(first_line) - len(first_line.lstrip())
    indent_len = len(indent)
    if first_line_indent >= indent_len:
        return code

    required_extra_indent = " " * (indent_len - first_line_indent)

    after_first_line = code_split[1:]
    return "\n".join(
        [first_line, *[f"{required_extra_indent}{line}" for line in after_first_line]]
    )


def strip_empty_lines(inp: str) -> str:
    """
    Strip empty lines

    More specifically, turn lines
    that only have whitespace into a single newline character.
    This is a workaround until we start formatting our Fortran
    as part of the generation process
    (https://gitlab.com/magicc/fgen/-/issues/26).

    Parameters
    ----------
    inp
        Input string

    Returns
    -------
        Input string, with all whitespace lines replaced by a single newline character.
    """
    out = [line if line.strip() else "\n" for line in inp.splitlines()]

    return "\n".join(out)


JINJA_ENV.filters["indent_based_on_first_line"] = indent_based_on_first_line
JINJA_ENV.filters["strip_empty_lines"] = strip_empty_lines


def get_template_in_directory(
    template_name: str, template_directory: Path, env: jinja2.Environment
) -> jinja2.Template:
    """
    Get template that is in the directory specified by ``file_dunder``

    Parameters
    ----------
    template_name
        Name of the template to get

    template_directory
        The directory from which to retrieve the template.

        If you want to retrieve a template in the same directory as the file
        that is calling this function,
        call ``get_template_in_directory(template_name, Path(__file__).parent), env)``.

    env
        Jinja2 environment to use when loading the template.
        This must be using a :obj:`jinja2.loaders.FileSystemLoader` loader.
        The loader must only have one search path.

    Returns
    -------
        Loaded template
    """
    if not isinstance(env.loader, jinja2.loaders.FileSystemLoader):
        raise NotImplementedError(f"{env.loader=}")

    if len(env.loader.searchpath) > 1:
        raise NotImplementedError(
            f"More than one search path: {env.loader.searchpath=}"
        )

    loader_searchpath = env.loader.searchpath[0]

    template_path = (template_directory / template_name).relative_to(loader_searchpath)

    return env.get_template(str(template_path))


def post_process_jinja_rendering(inp: str) -> str:
    """
    Post-process the result from Jinja

    This fixes things we can't work out how to fix with jinja

    Parameters
    ----------
    inp
        Result of formatting with jinja

    Returns
    -------
        Post-processed result
    """
    # Strip out two blank lines in a row
    two_blank_lines = "\n\n\n"
    one_blank_line = "\n\n"
    while two_blank_lines in inp:
        inp = inp.replace(two_blank_lines, one_blank_line)

    # Jinja trims the trailing line-break from the end of the rendered text,
    # add that back in too
    return inp + "\n"



"""
Data model of a Fortran module
"""

from __future__ import annotations

from attrs import define, field

from fgen.data_models.fortran_derived_type import FortranDerivedType
from fgen.data_models.module_requirement import ModuleRequirement


@define
class Module:
    """
    Data model of a Fortran module

    It is assumed that each module defines a single derived type.
    """

    name: str
    """Name of the module"""

    description: str
    """Description of the module"""

    provides: FortranDerivedType
    """Derived type that the module defines"""

    truncated_name: str = field()
    """
    If provided, the name to use when creating the Fortran wrapper module.

    If not supplied, we simply use the value of :attr:`name`.

    This can be used to avoid line length limits for the wrappers
    when we are wrapping modules with very long names.
    """

    requirements: list[ModuleRequirement] = field(factory=list)
    """
    Requirements from other modules

    Put another way, the modules that this module requires in order to run.
    """

    @truncated_name.default
    def truncated_name_default(self) -> str:
        """
        Default value for :attr:`truncated_name`
        """  # noqa: D401
        return self.name

    @property
    def wrapper_module_name(self) -> str:
        """
        Name for the Fortran wrapper module

        By default the wrapper module name is derived from the :attr:`name`,
        but, in some cases, the complete module name can be too long,
        leading to the wrapper module not being able to be built.
        In this case a :attr:`truncated_name` can be used
        to derive a shortened wrapper module name.
        Given that this wrapper module is only used by autogenerated code the
        readability of the name isn't as important as the full module.
        """
        return f"{self.truncated_name}_w"

    @property
    def manager_module_name(self) -> str:
        """
        Name for the Fortran manager module

        By default the manager module name is derived from the :attr:`name`,
        but, in some cases, the complete module name can be too long,
        leading to the manager module not being able to be built.
        In this case a :attr:`truncated_name` can be used
        to derive a shortened manager module name.
        Given that this manager module is only used by autogenerated code the
        readability of the name isn't as important as the full module.
        """
        return f"{self.truncated_name}_manager"

    def get_requirement_that_provides(self, provides: str) -> ModuleRequirement:
        """
        Get the module requirement that provides the specified type

        Parameters
        ----------
        provides
            Name of the provided type

        Returns
        -------
            Module requirement that provides this type

        Raises
        ------
        ValueError
            No module requirement is found that provides ``provides``
            is part of ``self.requirements``.
        """
        for requirement in self.requirements:
            if requirement.provides == provides:
                return requirement

        msg = f"No requirement provides {provides}. {self.requirements=}"
        raise ValueError(msg)



"""
Interface definition for wrapping strategies
"""
from __future__ import annotations

from collections.abc import Iterable
from typing import Optional, Protocol, Union

from fgen.data_models import Method, Module, MultiReturn, Value
from fgen.wrapping_strategies.passing_to_fortran_steps import (
    PassingToFortranSteps,
)
from fgen.wrapping_strategies.receiving_from_python_steps import (
    ReceivingFromPythonSteps,
)


class WrappingStrategyLike(Protocol):
    """
    Wrapping strategy interface
    """

    def get_fortran_wrapper_module_getter(
        self, value: Union[MultiReturn, Value]
    ) -> str:
        """
        Get the name of the getter in the Fortran wrapper module

        Parameters
        ----------
        value
            Value for which to get the getter

        Returns
        -------
            Name of the getter in the Fortran wrapper module
        """
        ...

    def get_fortran_wrapper_module_setter(
        self, value: Union[MultiReturn, Value]
    ) -> str:
        """
        Get the name of the setter in the Fortran wrapper module

        Parameters
        ----------
        value
            Value for which to get the getter

        Returns
        -------
            Name of the getter in the Fortran wrapper module
        """
        ...

    def get_fortran_wrapper_statement_declarations_getters_and_setters(
        self, value: Union[MultiReturn, Value]
    ) -> list[str]:
        """
        Get statement declarations for the getters/setters in the Fortran wrapper module

        Parameters
        ----------
        value
            Value for which to get the statement declarations

        Returns
        -------
            Statement declarations for the getters and setters
            in the Fortran wrapper module.
        """
        ...

    def get_fortran_wrapper_statement_declarations_for_returning_method_result_to_python(  # noqa: E501
        self, method: Method, prefix: str
    ) -> list[str]:
        """
        Get statement declarations for the returning a method's result to Python

        Parameters
        ----------
        method
            Method for which to get the statement declarations.

        prefix
            Prefix to use when naming the method in Fortran.

        Returns
        -------
            Statement declarations for the returning a method's result to Python.
        """
        ...

    def get_receiving_from_python_steps(
        self,
        value: Union[MultiReturn, Value],
        dynamic_unit: Optional[str] = None,
        providing_module: Optional[Module] = None,
        get_instance_callable_name: Optional[str] = None,
    ) -> ReceivingFromPythonSteps:
        """
        Get the steps required to receive the value from Python

        Parameters
        ----------
        value
            Value to receive from Python

        dynamic_unit
            The source of the dynamic unit, if this value has one.

        providing_module
            The value that provides ``value``'s type, if it is not intrinsic.

        get_instance_callable_name
            The name of the callable used to get instances of wrapped derived types
            (typically comes from
            :py:attr:`fgen.wrapper_building.fortran_wrapper_module.FortranWrapperModuleBuilder.shared`).

        Returns
        -------
            Steps required to receive the value from Python
        """
        ...

    def get_type_attribute_declaration_for_value_to_pass_to_original_fortran_module(
        self, value: Union[MultiReturn, Value]
    ) -> str:
        """
        Get the type attribute declaration for the value used to pass information to the original Fortran module

        This is not always the same as the type used by the original Fortran module,
        particularly when allocatable types are involved.

        For the definition of a type attribute declaration,
        see :py:mod:`fgen.fortran_parsing`.


        Parameters
        ----------
        value
            Value for which to get the type attribute declaration
            for the value to pass to the original Fortran.

        Returns
        -------
            Type attribute declaration
            for the value to use to pass ``value`` to the original Fortran.
        """  # noqa: E501
        ...

    def get_fortran_for_getter(
        self,
        value: Union[MultiReturn, Value],
        class_being_wrapped: str,
        value_manager_get_free_instance_number: Optional[str] = None,
        value_manager_get_instance: Optional[str] = None,
    ) -> str:
        """
        Get the Fortran for the wrapper module's getter.

        Parameters
        ----------
        value
            Value returned from the getter.

        class_being_wrapped
            Derived type being wrapped.

        value_manager_get_free_instance_number
            Callable which gets a free instance number
            for objects of the type of ``value``.

        value_manager_get_instance
            Callable which gets an instance for objects of the type of ``value``.

        Returns
        -------
            Fortran for the wrapper module's getter.
        """
        ...

    def get_fortran_for_method_returning_wrapped_type(  # noqa: PLR0913
        self,
        fortran_wrapper_module_callable: str,
        receiving_from_python_steps: ReceivingFromPythonSteps,
        return_value: Union[MultiReturn, Value],
        class_being_wrapped: str,
        method_name: str,
        return_value_manager_get_free_instance_number: Optional[str],
        return_value_manager_get_instance: Optional[str],
    ) -> str:
        """
        Get the Fortran for a method which returns the wrapped type

        Parameters
        ----------
        fortran_wrapper_module_callable
            The Fortran wrapper module's callable's name

        receiving_from_python_steps
            Steps to receive the callable's inputs from Python

        return_value
            The (original) Fortran module callable's return value

        class_being_wrapped
            The name of the Fortran derived type being wrapped

        method_name
            The name of the method being wrapped

        return_value_manager_get_free_instance_number
            Callable which gets a free instance number
            for objects of the type of ``return_value``.

        return_value_manager_get_instance
            Callable which gets an instance for objects of the type of ``return_value``.

        Returns
        -------
            Fortran for wrapping the method
        """
        ...

    def get_python_user_facing_name(self, value: Union[MultiReturn, Value]) -> str:
        """
        Get the name that Python users will use for this value

        For example, when naming attributes
        or using this value as an argument in a function/method.

        Parameters
        ----------
        value
            Name that Python users will use for this value

        Returns
        -------
            Name to use for the Python attribute
        """
        ...

    def get_python_post_verify_units_input_type_annotation(
        self, value: Union[MultiReturn, Value]
    ) -> str:
        """
        Get input type annotation, assuming the ``verify_units`` decorator is applied.

        The type annotation is for when ``value``
        is used as an input to a Python callable.

        Parameters
        ----------
        value
            Value for which to get the type annotation

        Returns
        -------
            Type annotation for the value,
            assuming the ``verify_units`` decorator is applied.
        """
        ...

    def get_python_return_type_annotation(
        self, value: Union[MultiReturn, Value]
    ) -> str:
        """
        Get the return type annotation for the Python attribute

        Parameters
        ----------
        value
            Value for which to get the return type annotation for the Python attribute

        Returns
        -------
            Return type annotation for the Python attribute
        """
        ...

    def get_python_argument_declaration_type_annotation(self, value: Value) -> str:
        """
        Get the type annotation to use in Python user-facing declarations

        Parameters
        ----------
        value
            Value for which to get the type annotation

        Returns
        -------
            Type annotation to use in the Python argument declaration
        """
        ...

    def get_python_getter_docstring(self, value: Value) -> str:
        """
        Get docstring for the Python getter for the value

        Parameters
        ----------
        value
            Value for which to get the Python getter docstring

        Returns
        -------
            Python getter docstring
        """
        ...

    def get_python_setter_docstring(self, value: Value) -> str:
        """
        Get docstring for the Python setter for the value

        Parameters
        ----------
        value
            Value for which to get the Python setter docstring

        Returns
        -------
            Python setter docstring
        """
        ...

    def generate_python_for_fortran_return_value_processing(
        self,
        value: Union[MultiReturn, Value],
        fortran_module_callable: str,
        fortran_module_callable_args: Iterable[str],
        dynamic_unit: Optional[str],
    ) -> str:
        """
        Generate the Python code that processes the value returned from Fortran

        Parameters
        ----------
        value
            Value for which to generate the Python code for processing
            the value returned from Fortran

        fortran_module_callable
            The Fortran module callable to call

        fortran_module_callable_args
            The arguments to pass to the Fortran module callable
            (sometimes these need to be supplemented, hence we pass them separately).

        dynamic_unit
            If provided, specifies where to retrieve the dynamic unit of ``value`` from.

        Returns
        -------
            Python code that calls the callable and processes the results
            into the expected Python type.
        """
        ...

    def get_passing_to_fortran_steps(
        self,
        value: Union[MultiReturn, Value],
        dynamic_unit: Optional[str],
    ) -> PassingToFortranSteps:
        """
        Get the steps required to pass this value to the Fortran

        Parameters
        ----------
        value
            Value to pass into Fortran

        dynamic_unit
            If provided, specifies the dynamic units to convert values to.

        Returns
        -------
            Steps required to pass the value into the Fortran
        """
        ...
